@@ -26,6 +26,7 @@ from typing import Type
 from typing import Optional
 from typing import Any
 from typing import Union
+from typing import Callable
 from typing import TYPE_CHECKING
 
 import time
@@ -36,6 +37,8 @@ import socket
 import random
 import binascii
 import datetime
+import traceback
+import atexit
 
 from libcloud.utils.py3 import b
 
@@ -996,6 +999,8 @@ class NodeDriver(BaseDriver):
                     timeout=SSH_CONNECT_TIMEOUT,  # type: int
                     max_tries=3,  # type: int
                     ssh_interface='public_ips',  # type: str
+                    at_exit_func=None,  # type: Callable
+                    wait_period=5,  # type: int
                     **create_node_kwargs):
         # type: (...) -> Node
         """
@@ -1085,6 +1090,29 @@ class NodeDriver(BaseDriver):
         :param ssh_interface: The interface to wait for. Default is
                                    'public_ips', other option is 'private_ips'.
         :type ssh_interface: ``str``
+
+        :param at_exit_func: Optional atexit handler function which will be
+                             registered and called with created node if user
+                             cancels the deploy process (e.g. CTRL+C), after
+                             the node has been created, but before the deploy
+                             process has finished.
+
+                             This method gets passed in two keyword arguments:
+
+                             - driver -> node driver in question
+                             - node -> created Node object
+
+                             Keep in mind that this function will only be
+                             called in such scenario. In case the method
+                             finishes (this includes throwing an exception),
+                             at exit handler function won't be called.
+        :type at_exit_func: ``func``
+
+        :param wait_period: How many seconds to wait between each iteration
+                            while waiting for node to transition into
+                            running state and have IP assigned. (default is 5)
+        :type wait_period: ``int``
+
         """
         if not libcloud.compute.ssh.have_paramiko:
             raise RuntimeError('paramiko is not installed. You can install ' +
@@ -1146,6 +1174,9 @@ class NodeDriver(BaseDriver):
             else:
                 raise e
 
+        if at_exit_func:
+            atexit.register(at_exit_func, driver=self, node=node)
+
         password = None
         if auth:
             if isinstance(auth, NodeAuthPassword):
@@ -1159,10 +1190,13 @@ class NodeDriver(BaseDriver):
         try:
             node, ip_addresses = self.wait_until_running(
                 nodes=[node],
-                wait_period=3,
+                wait_period=wait_period,
                 timeout=wait_timeout,
                 ssh_interface=ssh_interface)[0]
         except Exception as e:
+            if at_exit_func:
+                atexit.unregister(at_exit_func)
+
             raise DeploymentError(node=node, original_exception=e, driver=self)
 
         ssh_alternate_usernames = ssh_alternate_usernames or []
@@ -1190,9 +1224,15 @@ class NodeDriver(BaseDriver):
                 break
 
         if deploy_error is not None:
+            if at_exit_func:
+                atexit.unregister(at_exit_func)
+
             raise DeploymentError(node=node, original_exception=deploy_error,
                                   driver=self)
 
+        if at_exit_func:
+            atexit.unregister(at_exit_func)
+
         return node
 
     def reboot_node(self, node):
@@ -1566,7 +1606,7 @@ class NodeDriver(BaseDriver):
 
     def wait_until_running(self,
                            nodes,  # type: List[Node]
-                           wait_period=3,  # type: float
+                           wait_period=5,  # type: float
                            timeout=600,  # type: int
                            ssh_interface='public_ips',  # type: str
                            force_ipv4=True,  # type: bool
@@ -1758,7 +1798,13 @@ class NodeDriver(BaseDriver):
 
                 # Retry if a connection is refused, timeout occurred,
                 # or the connection fails due to failed authentication.
-                ssh_client.close()
+                try:
+                    ssh_client.close()
+                except Exception:
+                    # Exception on close() should not be fatal since client
+                    # socket might already be closed
+                    pass
+
                 time.sleep(wait_period)
                 continue
             else:
@@ -1836,9 +1882,28 @@ class NodeDriver(BaseDriver):
             except Exception as e:
                 tries += 1
 
+                if "ssh session not active" in str(e).lower():
+                    # Sometimes connection gets closed or disconnected half
+                    # way through for wahtever reason.
+                    # If this happens, we try to re-connect before
+                    # re-attempting to run the step.
+                    try:
+                        ssh_client.close()
+                    except Exception:
+                        # Non fatal since connection is most likely already
+                        # closed at this point
+                        pass
+
+                    timeout = (int(ssh_client.timeout) if ssh_client.timeout
+                               else 10)
+                    ssh_client = self._ssh_client_connect(
+                        ssh_client=ssh_client,
+                        timeout=timeout)
+
                 if tries >= max_tries:
-                    raise LibcloudError(value='Failed after %d tries: %s'
-                                        % (max_tries, str(e)), driver=self)
+                    tb = traceback.format_exc()
+                    raise LibcloudError(value='Failed after %d tries: %s.\n%s'
+                                        % (max_tries, str(e), tb), driver=self)
             else:
                 # Deployment succeeded
                 ssh_client.close()
@@ -1847,7 +1912,7 @@ class NodeDriver(BaseDriver):
         return node
 
     def _get_size_price(self, size_id):
-        # type: (str) -> float
+        # type: (str) -> Optional[float]
         """
         Return pricing information for the provided size id.
         """
