@@ -13,10 +13,15 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from typing import Dict
+from typing import Optional
+
 import base64
 import hmac
 import time
 from hashlib import sha1
+import os
+from datetime import datetime
 
 import libcloud.utils.py3
 
@@ -32,17 +37,19 @@ from libcloud.utils.py3 import httplib
 from libcloud.utils.py3 import urlquote
 from libcloud.utils.py3 import b
 from libcloud.utils.py3 import tostring
+from libcloud.utils.py3 import urlencode
 
 from libcloud.utils.xml import fixxpath, findtext
 from libcloud.utils.files import read_in_chunks
 from libcloud.common.types import InvalidCredsError, LibcloudError
 from libcloud.common.base import ConnectionUserAndKey, RawResponse
 from libcloud.common.aws import AWSBaseResponse, AWSDriver, \
-    AWSTokenConnection, SignedAWSConnection
+    AWSTokenConnection, SignedAWSConnection, UnsignedPayloadSentinel
 
 from libcloud.storage.base import Object, Container, StorageDriver
 from libcloud.storage.types import ContainerError
 from libcloud.storage.types import ContainerIsNotEmptyError
+from libcloud.storage.types import ContainerAlreadyExistsError
 from libcloud.storage.types import InvalidContainerNameError
 from libcloud.storage.types import ContainerDoesNotExistError
 from libcloud.storage.types import ObjectDoesNotExistError
@@ -55,6 +62,7 @@ S3_US_STANDARD_HOST = 's3.amazonaws.com'
 S3_US_EAST2_HOST = 's3-us-east-2.amazonaws.com'
 S3_US_WEST_HOST = 's3-us-west-1.amazonaws.com'
 S3_US_WEST_OREGON_HOST = 's3-us-west-2.amazonaws.com'
+S3_US_GOV_EAST_HOST = 's3-us-gov-east-1.amazonaws.com'
 S3_US_GOV_WEST_HOST = 's3-us-gov-west-1.amazonaws.com'
 S3_CN_NORTH_HOST = 's3.cn-north-1.amazonaws.com.cn'
 S3_CN_NORTHWEST_HOST = 's3.cn-northwest-1.amazonaws.com.cn'
@@ -78,6 +86,7 @@ REGION_TO_HOST_MAP = {
     'us-east-2': S3_US_EAST2_HOST,
     'us-west-1': S3_US_WEST_HOST,
     'us-west-2': S3_US_WEST_OREGON_HOST,
+    'us-gov-east-1': S3_US_GOV_EAST_HOST,
     'us-gov-west-1': S3_US_GOV_WEST_HOST,
     'cn-north-1': S3_CN_NORTH_HOST,
     'cn-northwest-1': S3_CN_NORTHWEST_HOST,
@@ -108,11 +117,17 @@ CHUNK_SIZE = 5 * 1024 * 1024
 # ex_iterate_multipart_uploads.
 RESPONSES_PER_REQUEST = 100
 
+S3_CDN_URL_DATETIME_FORMAT = '%Y%m%dT%H%M%SZ'
+S3_CDN_URL_DATE_FORMAT = '%Y%m%d'
+S3_CDN_URL_EXPIRY_HOURS = float(
+    os.getenv('LIBCLOUD_S3_CDN_URL_EXPIRY_HOURS', '24')
+)
+
 
 class S3Response(AWSBaseResponse):
     namespace = None
     valid_response_codes = [httplib.NOT_FOUND, httplib.CONFLICT,
-                            httplib.BAD_REQUEST]
+                            httplib.BAD_REQUEST, httplib.PARTIAL_CONTENT]
 
     def success(self):
         i = int(self.status)
@@ -286,38 +301,29 @@ class BaseS3StorageDriver(StorageDriver):
         raise LibcloudError('Unexpected status code: %s' % (response.status),
                             driver=self)
 
-    def list_container_objects(self, container, ex_prefix=None):
-        """
-        Return a list of objects for the given container.
-
-        :param container: Container instance.
-        :type container: :class:`Container`
-
-        :param ex_prefix: Only return objects starting with ex_prefix
-        :type ex_prefix: ``str``
-
-        :return: A list of Object instances.
-        :rtype: ``list`` of :class:`Object`
-        """
-        return list(self.iterate_container_objects(container,
-                                                   ex_prefix=ex_prefix))
-
-    def iterate_container_objects(self, container, ex_prefix=None):
+    def iterate_container_objects(self, container, prefix=None,
+                                  ex_prefix=None):
         """
         Return a generator of objects for the given container.
 
         :param container: Container instance
         :type container: :class:`Container`
 
+        :param prefix: Only return objects starting with prefix
+        :type prefix: ``str``
+
         :param ex_prefix: Only return objects starting with ex_prefix
         :type ex_prefix: ``str``
 
         :return: A generator of Object instances.
         :rtype: ``generator`` of :class:`Object`
         """
+        prefix = self._normalize_prefix_argument(prefix, ex_prefix)
+
         params = {}
-        if ex_prefix:
-            params['prefix'] = ex_prefix
+
+        if prefix:
+            params['prefix'] = prefix
 
         last_key = None
         exhausted = False
@@ -420,9 +426,15 @@ class BaseS3StorageDriver(StorageDriver):
             container = Container(name=container_name, extra=None, driver=self)
             return container
         elif response.status == httplib.CONFLICT:
+            if "BucketAlreadyOwnedByYou" in response.body:
+                raise ContainerAlreadyExistsError(
+                    value='Container with this name already exists. The name '
+                          'be unique among all the containers in the system.',
+                    container_name=container_name, driver=self)
+
             raise InvalidContainerNameError(
                 value='Container with this name already exists. The name must '
-                      'be unique among all the containers in the system',
+                      'be unique among all the containers in the system.',
                 container_name=container_name, driver=self)
         elif response.status == httplib.BAD_REQUEST:
             raise ContainerError(
@@ -478,8 +490,52 @@ class BaseS3StorageDriver(StorageDriver):
                              'chunk_size': chunk_size},
             success_status_code=httplib.OK)
 
+    def download_object_range(self, obj, destination_path, start_bytes,
+                              end_bytes=None, overwrite_existing=False,
+                              delete_on_failure=True):
+        self._validate_start_and_end_bytes(start_bytes=start_bytes,
+                                           end_bytes=end_bytes)
+
+        obj_path = self._get_object_path(obj.container, obj.name)
+
+        headers = {'Range': self._get_standard_range_str(start_bytes,
+                                                         end_bytes)}
+        response = self.connection.request(obj_path, method='GET',
+                                           headers=headers, raw=True)
+
+        return self._get_object(obj=obj, callback=self._save_object,
+                                response=response,
+                                callback_kwargs={
+                                    'obj': obj,
+                                    'response': response.response,
+                                    'destination_path': destination_path,
+                                    'overwrite_existing': overwrite_existing,
+                                    'delete_on_failure': delete_on_failure,
+                                    'partial_download': True},
+                                success_status_code=httplib.PARTIAL_CONTENT)
+
+    def download_object_range_as_stream(self, obj, start_bytes, end_bytes=None,
+                                        chunk_size=None):
+        self._validate_start_and_end_bytes(start_bytes=start_bytes,
+                                           end_bytes=end_bytes)
+
+        obj_path = self._get_object_path(obj.container, obj.name)
+
+        headers = {'Range': self._get_standard_range_str(start_bytes,
+                                                         end_bytes)}
+        response = self.connection.request(obj_path, method='GET',
+                                           headers=headers,
+                                           stream=True, raw=True)
+
+        return self._get_object(
+            obj=obj, callback=read_in_chunks,
+            response=response,
+            callback_kwargs={'iterator': response.iter_content(CHUNK_SIZE),
+                             'chunk_size': chunk_size},
+            success_status_code=httplib.PARTIAL_CONTENT)
+
     def upload_object(self, file_path, container, object_name, extra=None,
-                      verify_hash=True, ex_storage_class=None):
+                      verify_hash=True, headers=None, ex_storage_class=None):
         """
         @inherits: :class:`StorageDriver.upload_object`
 
@@ -489,6 +545,7 @@ class BaseS3StorageDriver(StorageDriver):
         return self._put_object(container=container, object_name=object_name,
                                 extra=extra, file_path=file_path,
                                 verify_hash=verify_hash,
+                                headers=headers,
                                 storage_class=ex_storage_class)
 
     def _initiate_multipart(self, container, object_name, headers=None):
@@ -671,7 +728,8 @@ class BaseS3StorageDriver(StorageDriver):
                                 (resp.status), driver=self)
 
     def upload_object_via_stream(self, iterator, container, object_name,
-                                 extra=None, ex_storage_class=None):
+                                 extra=None, headers=None,
+                                 ex_storage_class=None):
         """
         @inherits: :class:`StorageDriver.upload_object_via_stream`
 
@@ -691,10 +749,12 @@ class BaseS3StorageDriver(StorageDriver):
                                               extra=extra,
                                               stream=iterator,
                                               verify_hash=False,
+                                              headers=headers,
                                               storage_class=ex_storage_class)
         return self._put_object(container=container, object_name=object_name,
                                 extra=extra, method=method, query_args=params,
                                 stream=iterator, verify_hash=False,
+                                headers=headers,
                                 storage_class=ex_storage_class)
 
     def delete_object(self, obj):
@@ -808,13 +868,14 @@ class BaseS3StorageDriver(StorageDriver):
             self._abort_multipart(container, upload.key, upload.id)
 
     def _clean_object_name(self, name):
-        name = urlquote(name)
+        name = urlquote(name, safe='/~')
         return name
 
     def _put_object(self, container, object_name, method='PUT',
                     query_args=None, extra=None, file_path=None,
-                    stream=None, verify_hash=True, storage_class=None):
-        headers = {}
+                    stream=None, verify_hash=True, storage_class=None,
+                    headers=None):
+        headers = headers or {}
         extra = extra or {}
 
         headers.update(self._to_storage_class_headers(storage_class))
@@ -875,7 +936,7 @@ class BaseS3StorageDriver(StorageDriver):
                 driver=self)
 
     def _put_object_multipart(self, container, object_name, stream,
-                              extra=None, verify_hash=False,
+                              extra=None, verify_hash=False, headers=None,
                               storage_class=None):
         """
         Uploads an object using the S3 multipart algorithm.
@@ -895,13 +956,16 @@ class BaseS3StorageDriver(StorageDriver):
         :keyword extra: Additional options
         :type extra: ``dict``
 
+        :keyword headers: Additional headers
+        :type headers: ``dict``
+
         :keyword storage_class: The name of the S3 object's storage class
         :type extra: ``str``
 
         :return: The uploaded object
         :rtype: :class:`Object`
         """
-        headers = {}
+        headers = headers or {}
         extra = extra or {}
 
         headers.update(self._to_storage_class_headers(storage_class))
@@ -910,12 +974,8 @@ class BaseS3StorageDriver(StorageDriver):
         meta_data = extra.get('meta_data', None)
         acl = extra.get('acl', None)
 
-        if not content_type:
-            content_type, _ = libcloud.utils.files.guess_file_mime_type(
-                object_name)
-
-        if content_type:
-            headers['Content-Type'] = content_type
+        headers['Content-Type'] = self._determine_content_type(
+            content_type, object_name)
 
         if meta_data:
             for key, value in list(meta_data.items()):
@@ -990,6 +1050,15 @@ class BaseS3StorageDriver(StorageDriver):
 
         return container
 
+    def _get_content_length_from_headers(self,
+                                         headers: Dict[str, str]
+                                         ) -> Optional[int]:
+        """
+        Prase object size from the provided response headers.
+        """
+        content_length = headers.get("content-length", None)
+        return content_length
+
     def _headers_to_object(self, object_name, container, headers):
         hash = headers['etag'].replace('"', '')
         extra = {'content_type': headers['content-type'],
@@ -1006,7 +1075,13 @@ class BaseS3StorageDriver(StorageDriver):
             key = key.replace(self.http_vendor_prefix + '-meta-', '')
             meta_data[key] = value
 
-        obj = Object(name=object_name, size=headers['content-length'],
+        content_length = self._get_content_length_from_headers(headers=headers)
+
+        if content_length is None:
+            raise KeyError("Can not deduce object size from headers for "
+                           "object %s" % (object_name))
+
+        obj = Object(name=object_name, size=int(content_length),
                      hash=hash, extra=extra,
                      meta_data=meta_data,
                      container=container,
@@ -1073,6 +1148,67 @@ class S3StorageDriver(AWSDriver, BaseS3StorageDriver):
     def list_regions(self):
         return REGION_TO_HOST_MAP.keys()
 
+    def get_object_cdn_url(self, obj,
+                           ex_expiry=S3_CDN_URL_EXPIRY_HOURS):
+        """
+        Return a "presigned URL" for read-only access to object
+
+        AWS only - requires AWS signature V4 authenticaiton.
+
+        :param obj: Object instance.
+        :type  obj: :class:`Object`
+
+        :param ex_expiry: The number of hours after which the URL expires.
+                          Defaults to 24 hours or the value of the environment
+                          variable "LIBCLOUD_S3_STORAGE_CDN_URL_EXPIRY_HOURS",
+                          if set.
+        :type  ex_expiry: ``float``
+
+        :return: Presigned URL for the object.
+        :rtype: ``str``
+        """
+
+        # assemble data for the request we want to pre-sign
+        # see: https://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-query-string-auth.html # noqa
+        object_path = self._get_object_path(obj.container, obj.name)
+        now = datetime.utcnow()
+        duration_seconds = int(ex_expiry * 3600)
+        credparts = (
+            self.key,
+            now.strftime(S3_CDN_URL_DATE_FORMAT),
+            self.region,
+            's3',
+            'aws4_request')
+        params_to_sign = {
+            'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
+            'X-Amz-Credential': '/'.join(credparts),
+            'X-Amz-Date': now.strftime(S3_CDN_URL_DATETIME_FORMAT),
+            'X-Amz-Expires': duration_seconds,
+            'X-Amz-SignedHeaders': 'host'}
+        headers_to_sign = {'host': self.connection.host}
+
+        # generate signature for the pre-signed request
+        signature = self.connection.signer._get_signature(
+            params=params_to_sign,
+            headers=headers_to_sign,
+            dt=now,
+            method='GET',
+            path=object_path,
+            data=UnsignedPayloadSentinel
+        )
+
+        # Create final params for pre-signed URL
+        params = params_to_sign.copy()
+        params['X-Amz-Signature'] = signature
+
+        return '{scheme}://{host}:{port}{path}?{params}'.format(
+            scheme='https' if self.secure else 'http',
+            host=self.connection.host,
+            port=self.connection.port,
+            path=object_path,
+            params=urlencode(params),
+        )
+
 
 class S3USEast2Connection(S3SignatureV4Connection):
     host = S3_US_EAST2_HOST
@@ -1107,6 +1243,17 @@ class S3USWestOregonStorageDriver(S3StorageDriver):
     region_name = 'us-west-2'
 
 
+class S3USGovEastConnection(S3SignatureV4Connection):
+    host = S3_US_GOV_EAST_HOST
+
+
+class S3USGovEastStorageDriver(S3StorageDriver):
+    name = 'Amazon S3 (us-gov-east-1)'
+    connectionCls = S3USGovEastConnection
+    ex_location_name = 'us-gov-east-1'
+    region_name = 'us-gov-east-1'
+
+
 class S3USGovWestConnection(S3SignatureV4Connection):
     host = S3_US_GOV_WEST_HOST
 
